"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Granularity = void 0;
const moment_timezone_1 = __importDefault(require("moment-timezone"));
const shared_1 = require("@cubejs-backend/shared");
class Granularity {
    query;
    granularity;
    granularityInterval;
    queryTimezone;
    granularityOffset;
    origin;
    predefinedGranularity;
    constructor(query, timeDimension) {
        this.query = query;
        this.granularity = timeDimension.granularity;
        this.predefinedGranularity = (0, shared_1.isPredefinedGranularity)(this.granularity);
        this.queryTimezone = query.timezone;
        this.origin = moment_timezone_1.default.tz(query.timezone).startOf('year'); // Defaults to current year start
        if (this.predefinedGranularity) {
            this.granularityInterval = `1 ${this.granularity}`;
        }
        else {
            const customGranularity = this.query.cacheValue(['customGranularity', timeDimension.dimension, this.granularity], () => query.cubeEvaluator
                .resolveGranularity([...query.cubeEvaluator.parsePath('dimensions', timeDimension.dimension), 'granularities', this.granularity]));
            if (!customGranularity) {
                throw new Error(`Granularity "${timeDimension.granularity}" does not exist in dimension ${timeDimension.dimension}`);
            }
            this.granularityInterval = customGranularity.interval;
            if (customGranularity.origin) {
                this.origin = moment_timezone_1.default.tz(customGranularity.origin, query.timezone);
            }
            else if (customGranularity.offset) {
                this.granularityOffset = customGranularity.offset;
                this.origin = (0, shared_1.addInterval)(this.origin, (0, shared_1.parseSqlInterval)(customGranularity.offset));
            }
        }
    }
    isPredefined() {
        return this.predefinedGranularity;
    }
    /**
     * @returns origin date string in Query timezone
     */
    originLocalFormatted() {
        return this.origin.tz(this.queryTimezone).format('YYYY-MM-DDTHH:mm:ss.SSS');
    }
    /**
     * @returns origin date string in UTC timezone
     */
    originUtcFormatted() {
        return this.origin.clone().utc().format('YYYY-MM-DDTHH:mm:ss.SSSZ');
    }
    minGranularity() {
        if (this.predefinedGranularity) {
            return this.granularity;
        }
        if (this.granularityOffset) {
            return this.query.minGranularity(this.granularityFromInterval(), this.granularityFromOffset());
        }
        if (this.origin) {
            return this.query.minGranularity(this.granularityFromInterval(), this.query.granularityFor(this.origin.utc()));
        }
        return this.granularityFromInterval();
    }
    timeSeriesForInterval(dateRange, options = { timestampPrecision: 3 }) {
        if (this.predefinedGranularity) {
            return (0, shared_1.timeSeries)(this.granularity, dateRange, options);
        }
        // Interval range doesn't take timezone into account and operate in kinda local timezone,
        // but origin is treated as a timestamp in query timezone, so we pass it as the naive timestamp
        // to be in sync with date range during calculation.
        return (0, shared_1.timeSeriesFromCustomInterval)(this.granularityInterval, dateRange, (0, moment_timezone_1.default)(this.originLocalFormatted()), options);
    }
    resolvedGranularity() {
        if (this.predefinedGranularity) {
            return this.granularity;
        }
        return this.granularityFromInterval();
    }
    /**
     * Returns the smallest granularity for the granularityInterval
     */
    granularityFromInterval() {
        return this.granularityFromIntervalString(this.granularityInterval);
    }
    /**
     * Returns the smallest granularity for the granularityOffset
     */
    granularityFromOffset() {
        return this.granularityOffset ? this.granularityFromIntervalString(this.granularityOffset) : '';
    }
    /**
     * Returns the smallest granularity for the provided interval string
     * Interval may be presented as `1 year 2 months 3 weeks 4 days 5 hours 6 minutes 7 seconds
     * It is important to bubble up from the smallest, as this is used e.g. for minimum rollup granularity
     */
    granularityFromIntervalString(interval) {
        const intervalParsed = (0, shared_1.parseSqlInterval)(interval);
        const intervalKeys = Object.keys(intervalParsed);
        if (intervalKeys.length === 1) {
            return intervalKeys[0];
        }
        if (intervalParsed.second) {
            return 'second';
        }
        else if (intervalParsed.minute) {
            return 'minute';
        }
        else if (intervalParsed.hour) {
            return 'hour';
        }
        else if (intervalParsed.day) {
            return 'day';
        }
        else if (intervalParsed.week) {
            return 'day';
        }
        else if (intervalParsed.month) {
            return 'month';
        }
        else if (intervalParsed.quarter) { // Only quarter+years possible
            return 'month';
        }
        else /* if (intervalParsed.year) */ {
            return 'year';
        }
    }
    isNaturalAligned() {
        const intParsed = this.granularityInterval.split(' ');
        return !(intParsed.length !== 2 || intParsed[0] !== '1');
    }
}
exports.Granularity = Granularity;
//# sourceMappingURL=Granularity.js.map